@@ -82,7 +82,7 @@ module URI
         if args.kind_of?(Array)
           return self.build(args.collect{|x|
             if x.is_a?(String)
-              URI::RFC2396_PARSER.escape(x)
+              DEFAULT_PARSER.escape(x)
             else
               x
             end
@@ -91,7 +91,7 @@ module URI
           tmp = {}
           args.each do |key, value|
             tmp[key] = if value
-                URI::RFC2396_PARSER.escape(value)
+                DEFAULT_PARSER.escape(value)
               else
                 value
               end
@@ -186,18 +186,18 @@ module URI
 
       if arg_check
         self.scheme = scheme
+        self.userinfo = userinfo
         self.hostname = host
         self.port = port
-        self.userinfo = userinfo
         self.path = path
         self.query = query
         self.opaque = opaque
         self.fragment = fragment
       else
         self.set_scheme(scheme)
+        self.set_userinfo(userinfo)
         self.set_host(host)
         self.set_port(port)
-        self.set_userinfo(userinfo)
         self.set_path(path)
         self.query = query
         self.set_opaque(opaque)
@@ -393,7 +393,7 @@ module URI
     def check_user(v)
       if @opaque
         raise InvalidURIError,
-          "cannot set user with opaque"
+          "can not set user with opaque"
       end
 
       return v unless v
@@ -417,7 +417,7 @@ module URI
     def check_password(v, user = @user)
       if @opaque
         raise InvalidURIError,
-          "cannot set password with opaque"
+          "can not set password with opaque"
       end
       return v unless v
 
@@ -511,7 +511,7 @@ module URI
         user, password = split_userinfo(user)
       end
       @user     = user
-      @password = password
+      @password = password if password
 
       [@user, @password]
     end
@@ -522,7 +522,7 @@ module URI
     # See also URI::Generic.user=.
     #
     def set_user(v)
-      set_userinfo(v, nil)
+      set_userinfo(v, @password)
       v
     end
     protected :set_user
@@ -574,12 +574,6 @@ module URI
       @password
     end
 
-    # Returns the authority info (array of user, password, host and
-    # port), if any is set.  Or returns +nil+.
-    def authority
-      return @user, @password, @host, @port if @user || @password || @host || @port
-    end
-
     # Returns the user component after URI decoding.
     def decoded_user
       URI.decode_uri_component(@user) if @user
@@ -602,7 +596,7 @@ module URI
 
       if @opaque
         raise InvalidURIError,
-          "cannot set host with registry or opaque"
+          "can not set host with registry or opaque"
       elsif parser.regexp[:HOST] !~ v
         raise InvalidComponentError,
           "bad component(expected host component): #{v}"
@@ -621,13 +615,6 @@ module URI
     end
     protected :set_host
 
-    # Protected setter for the authority info (+user+, +password+, +host+
-    # and +port+).  If +port+ is +nil+, +default_port+ will be set.
-    #
-    protected def set_authority(user, password, host, port = nil)
-      @user, @password, @host, @port = user, password, host, port || self.default_port
-    end
-
     #
     # == Args
     #
@@ -652,7 +639,6 @@ module URI
     def host=(v)
       check_host(v)
       set_host(v)
-      set_userinfo(nil)
       v
     end
 
@@ -699,7 +685,7 @@ module URI
 
       if @opaque
         raise InvalidURIError,
-          "cannot set port with registry or opaque"
+          "can not set port with registry or opaque"
       elsif !v.kind_of?(Integer) && parser.regexp[:PORT] !~ v
         raise InvalidComponentError,
           "bad component(expected port component): #{v.inspect}"
@@ -743,22 +729,21 @@ module URI
     def port=(v)
       check_port(v)
       set_port(v)
-      set_userinfo(nil)
       port
     end
 
     def check_registry(v) # :nodoc:
-      raise InvalidURIError, "cannot set registry"
+      raise InvalidURIError, "can not set registry"
     end
     private :check_registry
 
-    def set_registry(v) # :nodoc:
-      raise InvalidURIError, "cannot set registry"
+    def set_registry(v) #:nodoc:
+      raise InvalidURIError, "can not set registry"
     end
     protected :set_registry
 
-    def registry=(v) # :nodoc:
-      raise InvalidURIError, "cannot set registry"
+    def registry=(v)
+      raise InvalidURIError, "can not set registry"
     end
 
     #
@@ -881,7 +866,7 @@ module URI
       # hier_part     = ( net_path | abs_path ) [ "?" query ]
       if @host || @port || @user || @path  # userinfo = @user + ':' + @password
         raise InvalidURIError,
-          "cannot set opaque with host, port, userinfo or path"
+          "can not set opaque with host, port, userinfo or path"
       elsif v && parser.regexp[:OPAQUE] !~ v
         raise InvalidComponentError,
           "bad component(expected opaque component): #{v}"
@@ -960,7 +945,7 @@ module URI
     # == Description
     #
     # URI has components listed in order of decreasing significance from left to right,
-    # see RFC3986 https://www.rfc-editor.org/rfc/rfc3986 1.2.3.
+    # see RFC3986 https://tools.ietf.org/html/rfc3986 1.2.3.
     #
     # == Usage
     #
@@ -1136,7 +1121,7 @@ module URI
 
       base = self.dup
 
-      authority = rel.authority
+      authority = rel.userinfo || rel.host || rel.port
 
       # RFC2396, Section 5.2, 2)
       if (rel.path.nil? || rel.path.empty?) && !authority && !rel.query
@@ -1148,14 +1133,17 @@ module URI
       base.fragment=(nil)
 
       # RFC2396, Section 5.2, 4)
-      if authority
-        base.set_authority(*authority)
-        base.set_path(rel.path)
-      elsif base.path && rel.path
-        base.set_path(merge_path(base.path, rel.path))
+      if !authority
+        base.set_path(merge_path(base.path, rel.path)) if base.path && rel.path
+      else
+        # RFC2396, Section 5.2, 4)
+        base.set_path(rel.path) if rel.path
       end
 
       # RFC2396, Section 5.2, 7)
+      base.set_userinfo(rel.userinfo) if rel.userinfo
+      base.set_host(rel.host)         if rel.host
+      base.set_port(rel.port)         if rel.port
       base.query = rel.query       if rel.query
       base.fragment=(rel.fragment) if rel.fragment
 
@@ -1247,7 +1235,7 @@ module URI
         return rel, rel
       end
 
-      # you can modify `rel', but cannot `oth'.
+      # you can modify `rel', but can not `oth'.
       return oth, rel
     end
     private :route_from0
@@ -1272,7 +1260,7 @@ module URI
     #   #=> #<URI::Generic /main.rbx?page=1>
     #
     def route_from(oth)
-      # you can modify `rel', but cannot `oth'.
+      # you can modify `rel', but can not `oth'.
       begin
         oth, rel = route_from0(oth)
       rescue
@@ -1376,9 +1364,6 @@ module URI
           str << ':'
           str << @port.to_s
         end
-        if (@host || @port) && !@path.empty? && !@path.start_with?('/')
-          str << '/'
-        end
         str << @path
         if @query
           str << '?'
@@ -1404,18 +1389,29 @@ module URI
       end
     end
 
-    # Returns the hash value.
     def hash
       self.component_ary.hash
     end
 
-    # Compares with _oth_ for Hash.
     def eql?(oth)
       self.class == oth.class &&
       parser == oth.parser &&
       self.component_ary.eql?(oth.component_ary)
     end
 
+=begin
+
+--- URI::Generic#===(oth)
+
+=end
+#    def ===(oth)
+#      raise NotImplementedError
+#    end
+
+=begin
+=end
+
+
     # Returns an Array of the components defined from the COMPONENT Array.
     def component_ary
       component.collect do |x|
@@ -1452,7 +1448,7 @@ module URI
       end
     end
 
-    def inspect # :nodoc:
+    def inspect
       "#<#{self.class} #{self}>"
     end
 
